(ns clojureDemo.VisionCore)
  
(import '(java.awt Rectangle Robot Toolkit) )
(import '(java.awt.image BufferedImage) )
(import '(java.awt Graphics2D Panel))
(import '(java.io File) )
(import '(javax.imageio ImageIO) )
(import '(javax.swing JFrame))
(import '(org.apache.commons.io FileUtils))
(import clojure.lang.LazySeq)
(import '(name.audet.samuel.javacv.jna highgui cv cxcore))
(import '(name.audet.samuel.javacv CanvasFrame))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage))
(import '(name.audet.samuel.javacv.jna highgui$CvCapture$PointerByReference))
(import '(name.audet.samuel.javacv.jna highgui$CvVideoWriter$PointerByReference))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage$PointerByReference))
(import '(name.audet.samuel.javacv.jna cxcore$IplImage))
(import '(name.audet.samuel.javacv JavaCvErrorCallback))
(.redirectError (JavaCvErrorCallback.))

(use 'clojure.contrib.repl-utils)

(def -inf Double/NEGATIVE_INFINITY)
(def inf  Double/POSITIVE_INFINITY)



(def lian (File. "/home/r/Desktop/source-videos/lian1.mpeg"))
(def look (File. "/home/r/Desktop/source-videos/dramatic_look.flv"))

(def target (File. "/home/r/Desktop/output-vision/"))


;this is still a work in progress, I'll come back to it later when I understand 
;jna more thoroughly.  the important abstraction here is 
;video-seq, which gives a lazy sequence of Intel Image Processing library images.

(defn naturals [] (iterate inc 0))

(def ext "jpg") 
;see below for the rationale for this choice of extention.

(def cache-location "/home/r/Desktop/vision-cache/")

(defn close-capture
  [capture]
  (highgui/cvReleaseCapture (highgui$CvCapture$PointerByReference. capture)))

(defn close-writer
  [writer] (highgui/cvReleaseVideoWriter (highgui$CvVideoWriter$PointerByReference. writer))) 

(defn- cache-path
  [video]
  (File. cache-location (.getName video)))

(defn- already-cached
  "this is the simplest and most retarded way to do it"
  [video]
  (.exists (cache-path video)))

(defn- write-frame 
  [capture target-dir n]
  (let [image (highgui/cvQueryFrame capture)]
    (if (nil? image) false
	(highgui/cvSaveImage (str (File. target-dir (str n "."  ext))) image))))

(defn- write-frame-bad
  [capture target-dir n]
  (println (str "saving frame: " n))
  (let [image (highgui/cvQueryFrame capture)]
    (if (nil? image) false
	( ImageIO/write (.getBufferedImage image) ext (File. target-dir (str n "."  ext))))))

(defn- write-frames
  [video target-dir]
  (let [capture (highgui/cvCreateFileCapture (.getPath video))]
   (dorun 
    (for [n (naturals) :while (write-frame capture target-dir n) ] nil ))
    (highgui/cvReleaseCapture (highgui$CvCapture$PointerByReference. capture))))

(defn- cache-frames
  [cache-location video]
  (time 
   (do
     (println "\"caching entire video structure... this will take a while... go get a snack or something :)\"")
     (FileUtils/deleteDirectory (cache-path video))
     (FileUtils/forceMkdir (cache-path video))
     (write-frames video  (cache-path video)))))

(defn cache
  [video]
  (if (already-cached video) nil (cache-frames cache-location video)))
 
(defn video-len 
  [video] (cache video)
  (alength (.list (cache-path video))))
(def video-len (memoize video-len))
 
(defn video-data
  "since the opencv version is so absolutely unreliable..."
  [video]
  (let
      [capture (highgui/cvCreateFileCapture (.getPath video))
       info {:length      (video-len video)
	     :width       (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FRAME_WIDTH)
	     :height      (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FRAME_HEIGHT)
	     :fps         (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FPS)
	     :codec       (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FOURCC)}]
    (close-capture capture)
    info))
(def video-data (memoize video-data))

(defn- video-frame
  [video n] 
  (cache video)
  (let 
      [c++-managed  (highgui/cvLoadImage (str (File. (cache-path video) (str n "."  ext))) highgui/CV_LOAD_IMAGE_COLOR)
       jvm-managed (.clone c++-managed)] 
	 ;this bit with the cloning is so I can deal with Garbage Collection once and for all.
	 ;the cpp-managed image must be manually Garbage Collected, but it's clone is managed by 
         ;the JVM's Garbage Collector.  By getting rid of the c++ part right here and now, no 
         ;other function has to worry about manual garbage collection ever again.
         ;Unfortunately, this doesn't seem to work for certain types of files. It's not file-size
         ;which is the issue, but something involving the image header.
    (cxcore/cvReleaseImage (.pointerByReference c++-managed))
    jvm-managed
))
  

(defn- video-frame-buffered
  "takes one frame from a video in constant time"
  [video n]
  (cache video)
  (ImageIO/read (File. (cache-path video) (str n "."  ext))))


(defn- dumb-write 
  [video n writer]
  (let 
      [c++-managed  (highgui/cvLoadImage (str (File. (cache-path video) (str n ext))) highgui/CV_LOAD_IMAGE_COLOR)]
    (highgui/cvWriteFrame writer  c++-managed)
    (cxcore/cvReleaseImage (cxcore$IplImage$PointerByReference. c++-managed))))

(defn video-seq
  "makes a lazy sequence of IPL images"
  ;additionally, I want to pass metadata around with the sequence.
  [video] (cache video)
  (map #(video-frame video %) (range (video-len video))))






(comment

; I initially decided to use .sr because it loads the fastest out of all the 
; formats opencv supports, under a simple benchmark of reading/writing
; a blank file of each  type 100 times.

  ;I just kept changing the file extention at the REPL to generate these times.     
  (def file "test.tiff")
  (do 
    (time (dotimes [_ 100] (highgui/cvSaveImage (str cache-location file) ipl)))
    (time (dotimes [_ 100]  (highgui/cvLoadImage (str cache-location file)))))

  ;      Write              Read
  (jpg  4404.000955 msecs 3397.8564   msecs)
  (jpeg 4376.138853 msecs 3482.990118 msecs)
  (jpeg 4253.721501 msecs 3414.004122 msecs)
  (bmp  3488.281695 msecs  786.883035 msecs)
  (dib  3589.010247 msecs  685.681985 msecs)
  (jpe  4288.541679 msecs 3359.819425 msecs)
  (png 10127.648557 msecs 3786.184994 msecs)
  (pbm  3880.794141 msecs  917.737667 msecs)
  (pgm  3879.710445 msecs  894.78237  msecs)
  (ppm  3938.319148 msecs 1014.412766 msecs)
  (sr   3510.893891 msecs  676.502596 msecs)
  (dib  3434.654784 msecs  737.495844 msecs)
  (bmp  3354.956726 msecs  783.353025 msecs)
  (ras  3351.400751 msecs  722.548007 msecs)
  (tiff 3657.893326 msecs 1361.576798 msecs)
  (tif  3594.753736 msecs 1254.568533 msecs)

;Ah, but now it's time for some more tests.
;I started using
(def ext ".sr")
;, and an empty cache, and ran
(cache lian)
"caching entire video structure... this will take a while... go get a snack or something :)"
"Elapsed time: 56486.816728 msecs"
(time (dorun (video-seq lian)))
"Elapsed time: 120515.66221 msecs"
(time (dorun (video-seq lian)))
"Elapsed time: 122867.82989 msecs" ;good agreement with times

;*erased vision cache with*
;*rm -rf ~/Desktop/vision-cache *
(def ext ".bmp")
(cache lian)
"Elapsed time: 59613.624691 msecs"
(time (dorun (video-seq lian)))
"Elapsed time: 123850.390784 msecs"

;same process except with
(def ext ".jpg")
(cache lian)
"Elapsed time: 139964.031921 msecs"
(time (dorun (video-seq lian)))
"Elapsed time: 127740.50204 msecs"

;I find this quite shocking --- the jpg's do take longer to cache,
;but the processing time is almost the same!

;since lian is 434 MB as a bunch of jpg files, and 3.7 GB as .sr files,
;I'll go with the jpgs.


;; writing files

"JPG"
(time (write-video sarah (File. "/home/r/Desktop/clojure2.avi")))
"Elapsed time: 371541.024455 msecs"

"BMP"
(time (write-video sarah (File. "/home/r/Desktop/clojure3.avi")))
"Elapsed time: 382568.502361 msecs"

)
